package com.falcon.core.binding;

import com.falcon.core.binding.beans.*;
import com.falcon.core.binding.beans.Binding.DataTransfer;
import com.falcon.core.binding.beans.Method.Argument;
import com.falcon.core.exception.FalconValidationException;
import com.falcon.core.exception.NoDataTransferException;
import com.falcon.core.util.OutObjectHelper;
import org.apache.log4j.Logger;

import java.util.*;

/**
 * This class adds value to the jaxb beans that get generated. It allows
 * searching of datatransfer through a number of methods. It get created when
 * <code>TwineLoader.getXXXBindings</code> gets called. <br>
 * Date: 24-Jan-2005 Within Package: com.twine.core.binding
 *
 * @author Charles Lee
 */

public class BindingConfig implements BeanConfig {

    private Map<String, DataTransfer> dataTransferMap;
    private Map<String, DataTransfer> ejbMap;
    private Map<String, DataTransfer> allDtMap;
    private Map<String, List<DataTransfer>> depMap;
    private Map<String, OutObjectHelper> outObjMap;
    private transient Logger log = Logger.getLogger(this.getClass().getName());
    private String id;

    private BindingConfig() {
    }

    public String getId() {
        return id;
    }

    public BindingConfig(Falcon twine) throws FalconValidationException {
        loadBindingObjects(twine);
    }

    /**
     * Gets the list of dataTransfer objects for a pagetype <br>
     *
     * @return ArrayList of dataTransfer Objects.
     */
    public Map<String, DataTransfer> getAllPojos() {
        return dataTransferMap;
    }

    public DataTransfer getPojoByName(String name) throws NoDataTransferException {
        log.debug("get rntDt: " + name);
        DataTransfer rntDt;

        if (dataTransferMap.get(name) != null) {
            rntDt = dataTransferMap.get(name);
        } else {
            throw new NoDataTransferException("pojo " + name + " don't exist");
        }

        return rntDt;
    }

    public DataTransfer getPojoByClassObjectName(String classObj) throws NoDataTransferException {
        log.debug("get dt by class object: " + classObj);
        DataTransfer rntDt = null;
        boolean found = false;

        for (DataTransfer dt : dataTransferMap.values()) {
            log.debug("dtname: " + dt.getClassObject());
            if (dt.getClassObject().equals(classObj)) {
                rntDt = dt;
                found = true;
                break;
            }
        }

        if (!found)
            throw new NoDataTransferException("pojo Object " + classObj + " don't exist");
        return rntDt;
    }

    /**
     * Gets a list of dependent dataTransfers from the current dt name.
     * <b>Note Depedant dt that conatins a arg type of output will not be inculded</b>
     * <br>
     *
     * @param dtName
     * @return
     * @throws NoDataTransferException
     */
    public List<DataTransfer> getDependentDataTransfer(String dtName)
            throws NoDataTransferException {
        DataTransfer dt = getDtByName(dtName);
        List<DataTransfer> dependList = new ArrayList<DataTransfer>();
        Method meth;
        Method.Argument arg;

        for (ListIterator<Method> iter = dt.getMethod().listIterator(); iter.hasNext();) {
            meth = iter.next();
            for (ListIterator<Method.Argument> iterArg = meth.getArgument()
                    .listIterator(); iterArg.hasNext();) {

                arg = iterArg.next();
                if ((arg.getDtRef() != null) & (arg.getType() != ArgType.OUTPUT)) {
                    try {
                        dependList.add(getDtByName(arg.getDtRef()));
                    } catch (NoDataTransferException e) {
                        log.error("No dependent dt " + arg.getDtRef() + " found in binding Confif"
                                + this.id, e.fillInStackTrace());
                    }
                }
            }
        }

        //getting JPA dependants
        if (dt.getObjectStory() != null) {
            if (dt.getObjectStory().getJPA() != null) {
                if (dt.getObjectStory().getJPA().getParameters() != null) {
                    for (ObjectStory.JPA.Parameters parameters : dt.getObjectStory().getJPA().getParameters()) {
                        if (parameters.getDtRef() != null) dependList.add(getDtByName(parameters.getDtRef()));
                    }
                }
            }
        }

        return dependList;
    }

    /**
     * Returns a whole map of dependednt dts for the whole BindingConfig
     *
     * @return
     * @{link java.util.Map}
     */
    public Map<String, List<DataTransfer>> getDependentMap() {

        return depMap;
    }

    protected final Map<String, List<DataTransfer>> buildDepMap() {
        depMap = new HashMap<String, List<DataTransfer>>();
        for (String s : dataTransferMap.keySet()) {
            try {
                if (!getDependentDataTransfer(s).isEmpty()) {
                    depMap.put(s, getDependentDataTransfer(s));
                }
            } catch (NoDataTransferException e) {
                log.error("No dt found", e.fillInStackTrace());
            }
        }
        return depMap;
    }

    /**
     * Gets the list of ejb objects <br>
     *
     * @return Map of ejb Objects<Ejb name, Ejb dt object>.
     */
    public Map<String, DataTransfer> getAllEjbs() {

        return ejbMap;
    }

    public DataTransfer getEjbByName(String name) throws NoDataTransferException {
        log.debug("looking for ejb of: " + name);
        DataTransfer rntDt;

        if (ejbMap.get(name) != null) {
            rntDt = ejbMap.get(name);
        } else {
            throw new NoDataTransferException("EJB " + name + " don't exist");
        }

        return rntDt;
    }

//    public DataTransfer getEjbByJndi(String jndi) throws NoDataTransferException {
//        log.debug("looking for ejb of: " + jndi);
//        DataTransfer ejb = null;
//        boolean found = false;
//
//        for (DataTransfer dt : ejbMap.values()) {
//            if (dt.getJndi().equals(jndi)) {
//                ejb = dt;
//                found = true;
//                break;
//            }
//        }
//
//        if (!found)
//            throw new NoDataTransferException("EJB jndi " + jndi + " don't exist");
//
//        return ejb;
//    }

    public Map<String, DataTransfer> getAllDt() {
        return allDtMap;
    }

    public DataTransfer getDtByName(String name) throws NoDataTransferException {
        DataTransfer rntDt;

        if (allDtMap.get(name) != null) {
            rntDt = allDtMap.get(name);
        } else {
            throw new NoDataTransferException("dt " + name + " don't exist");
        }
        return rntDt;
    }

    public OutObjectHelper getOutPutDependantDt(String outputDt) {

        if (outObjMap.get(outputDt) != null) {
            return outObjMap.get(outputDt);
        }

        return null;
    }

    public void loadBindingObjects(Falcon twine) throws FalconValidationException {
        this.id = twine.getId();
        this.ejbMap = new HashMap<String, DataTransfer>();
        this.dataTransferMap = new HashMap<String, DataTransfer>();
        this.allDtMap = new HashMap<String, DataTransfer>();
        this.outObjMap = new HashMap<String, OutObjectHelper>();
        OutObjectHelper outHelper;

        for (DataTransfer dt : twine.getBinding().getDataTransfer()) {
            if (dt.getType() == DataTransferType.POJO || dt.getType() == DataTransferType.JPA) {
                if (this.dataTransferMap.get(dt.getName()) == null) {
                    this.dataTransferMap.put(dt.getName(), dt);
                    this.allDtMap.put(dt.getName(), dt);
                } else {
                    throw new FalconValidationException("Dt " + dt.getName() + " already exist");
                }
//            } else if (dt.getType() == DataTransferType.EJB) {
//                if (this.ejbMap.get(dt.getName()) == null) {
//                    this.ejbMap.put(dt.getName(), dt);
//                    this.allDtMap.put(dt.getName(), dt);
//                } else {
//                    throw new FalconValidationException("EJB " + dt.getName() + " already exist");
//                }

            }

            for (Method meth : dt.getMethod()) {
                for (Argument arg : meth.getArgument()) {
                    if (arg.getType().value().equals("output") & arg.getDtRef() != null) {
                        if (outObjMap.get(arg.getDtRef()) == null) {
                            outHelper = new OutObjectHelper();
                            outHelper.setFromDt(dt);
                            outHelper.setFromMethod(meth.getName());
                            outObjMap.put(arg.getDtRef(), outHelper);
                        } else {
                            throw new FalconValidationException(arg.getDtRef()
                                    + " is already referenced by dt "
                                    + outObjMap.get(arg.getDtRef()).getFromDt().getName() +
                                    " in method " + outObjMap.get(arg.getDtRef()).getFromMethod());
                        }
                    }
                }
            }
        }

        buildDepMap();
    }

}
